# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Description
#
#   Private functions.
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

################################################################################
#                                    Macros
################################################################################

"""
    @parse_value(T, str, line_num)

Parse the string `str` using the type `T`. If it is not succeeded, then throw an
error indicating the line `line_num` with the problem.

"""
macro parse_value(T, str, line_num)
    quote
        local _T = $(esc(T))
        local _str = $(esc(str))
        local _line_num = $(esc(line_num))

        try
            parse(_T, _str)
        catch e
            throw(ErrorException("The TLE file is not valid (error in line $_line_num): $(e.msg)."))
        end
    end
end

################################################################################
#                                  Functions
################################################################################

function _get_mant_exp(n)
    if abs(n) < 1e-20
        return 0., 0
    end

    exp  = floor(Int, log10(n)) + 1
    mant = n/10.0^exp

    return mant,exp
end

function _parse_tle(io::IO, verify_checksum::Bool = true)
    # State machine to read the TLE. It has three possible states:
    #
    #   :name -> Satellite name.
    #   :l1   -> Line 1.
    #   :l2   -> Line 2.
    #
    # The transitions are:
    #
    #   :name --> :l1 --> :l2
    #     ^                |
    #     |                |
    #      ----------------
    #
    state = :name

    # Auxiliary variables that will store the TLE.
    name = ""
    sat_num = 0
    classification = Char(0)
    int_designator = ""
    epoch_year = 0
    epoch_day = 0.0
    epoch = 0.0
    dn_o2 = 0.0
    ddn_o6 = 0.0
    bstar = 0.0
    elem_set_number = 0
    checksum_l1 = 0
    i = 0.0
    Ω = 0.0
    e = 0.0
    ω = 0.0
    M = 0.0
    n = 0.0
    rev_num = 0
    checksum_l2 = 0

    # Output array with the TLEs found in the file.
    tles = Array{TLE}(undef,0)

    line_num = 0

    skip_line_read = false

    line = nothing

    while !eof(io)
        # Read the current line, strip white spaces, and skip if it is blank or
        # is a comment.
        if !skip_line_read
            line = strip(readline(io))
            line_num += 1
            (isempty(line))  && continue
            (line[1] == '#') && continue
        else
            skip_line_read = false
        end

        # Check the state of the reading.
        if state == :name
            # Check if the line seems to be the first line of the TLE. In this
            # case, maybe the user has not provided a name. Then, change the
            # state to `:l1`, and skip the line reading in the next loop.
            if (line[1:2] == "1 ") && (length(line) == 69)
                name = "UNDEFINED"
                state = :l1
                skip_line_read  = true
                continue
            end

            # Otherwise, if the line is not blank, then it must be the name of
            # the satellite.
            #
            # NOTE: The name should not be bigger than 24 characters. However,
            # we will not check this here.

            # Copy the name and change to state to wait for the 1st line.
            name  = line
            state = :l1

        # TLE Line 1
        # ==========
        elseif state == :l1
            # The next non-blank line must be the first line of the TLE.
            # Otherwise, the file is not valid.

            # The first line must start with "1 " and have 69 characters.
            if (line[1:2] != "1 ") || (length(line) != 69)
                throw(ErrorException("The TLE file is not valid (error in line $line_num): This is not a valid 1st line."))
            end

            # Verify the checksum.
            # ====================

            checksum_l1   = @parse_value(Int, line[69], line_num)
            checksum_line = compute_checksum(line[1:end-1])

            if (verify_checksum) && (checksum_l1 != checksum_line)
                throw(ErrorException("The TLE file is not valid (error in line $line_num): Expected checksum: $checksum_line, line checksum: $checksum_l1."))
            end

            # Satellite number and classification
            # ===================================
            sat_num        = @parse_value(Int, line[3:7], line_num)
            classification = Char(line[8])

            # International designator
            # ========================
            int_designator = line[10:17]

            # Epoch
            # =====
            epoch_year = @parse_value(Int,     line[19:20], line_num)
            epoch_day  = @parse_value(Float64, line[21:32], line_num)

            # Convert the TLE year and date to JD.
            # ------------------------------------

            # If the two-number year is higher than 75, then consider it in the
            # past (e.g. 50 = 2050, but 81 = 1981).
            if epoch_year > 75
                epoch = datetime2julian(DateTime(1900 + epoch_year, 1, 1, 0, 0, 0)) - 1 +
                        epoch_day
            else
                epoch = datetime2julian(DateTime(2000 + epoch_year, 1, 1, 0, 0, 0)) - 1 +
                        epoch_day
            end

            # Mean motion derivatives
            # =======================
            dn_o2  = @parse_value(Float64, line[34:43], line_num)

            aux = ( (line[45] == ' ') ? "+." : line[45:45] * "." )*line[46:50]

            ddn_o6_dec = @parse_value(Float64, aux, line_num)
            ddn_o6_exp = @parse_value(Float64, line[51:52], line_num)
            ddn_o6 = ddn_o6_dec*10^ddn_o6_exp

            # BSTAR
            # =====

            aux = ( (line[54] == ' ') ? "+." : line[54:54] * "." )*line[55:59]

            bstar_dec = @parse_value(Float64, aux, line_num)
            bstar_exp = @parse_value(Float64, line[60:61], line_num)
            bstar = bstar_dec*10^bstar_exp

            # Ephemeris type
            # ==============
            (line[63] != '0' && line[63] != ' ') &&
                warn("Warning in TLE file (line $line_num): Ephemeris type should be 0!")

            # Element number
            # ==============
            elem_set_number = @parse_value(Int, line[65:68], line_num)

            # Change the state to wait for the line 2.
            state = :l2

        # TLE Line 2
        # ==========
        elseif state == :l2
            # The next non-blank line must be the second line of the TLE.
            # Otherwise, the file is not valid.

            # The second line must start with "2 " and have 69 characters.
            if (line[1:2] != "2 ") || (length(line) != 69)
                throw(ErrorException("The TLE file is not valid (error in line $line_num): This is not a valid 2st line."))
            end

            # Verify the checksum.
            # ====================

            checksum_l2   = @parse_value(Int, line[69], line_num)
            checksum_line = compute_checksum(line[1:end-1])

            if (verify_checksum) && (checksum_l2 != checksum_line)
                throw(ErrorException("The TLE file is not valid (error in line $line_num): Expected checksum: $checksum_line, line checksum: $checksum_l2."))
            end

            # Check satellite number with the one in the first line.
            # ======================================================
            sat_num_l2 = @parse_value(Int, line[3:7], line_num)

            if sat_num_l2 != sat_num
                throw(ErrorException("The TLE file is not valid (error in line $line_num): Satellite number in line 2 is not equal to that in line 1."))
            end

            # Inclination
            # ===========
            i = @parse_value(Float64, line[9:16], line_num)

            # RAAN
            # ====
            Ω = @parse_value(Float64, line[18:25], line_num)

            # Eccentricity
            # ============
            e = @parse_value(Float64, "." * line[27:33], line_num)

            # Argument of perigee
            # ===================
            ω = @parse_value(Float64, line[35:42], line_num)

            # Mean anomaly
            # ============
            M = @parse_value(Float64, line[44:51], line_num)

            # Mean motion
            # ===========
            n = @parse_value(Float64, line[53:63], line_num)

            # Revolution number at epoch
            # ==========================
            rev_num = @parse_value(Int, line[64:68], line_num)

            # Now that we have all the information, we can store our TLE in the
            # output array.
            push!(tles,
                  TLE(name,
                      sat_num,
                      classification,
                      int_designator,
                      epoch_year,
                      epoch_day,
                      epoch,
                      dn_o2,
                      ddn_o6,
                      bstar,
                      elem_set_number,
                      checksum_l1,
                      i,
                      Ω,
                      e,
                      ω,
                      M,
                      n,
                      rev_num,
                      checksum_l2))

            # Change the state to wait for another TLE.
            state = :name
        end
    end

    # If the final state is not :name, then we have an incomplete TLE. Thus,
    # throw and exception because the file is not valid.
    if state != :name
        throw(ErrorException("The TLE file is not valid.\nThere is an incomplete TLE."))
    end

    tles
end

"""
    compute_checksum(str::AbstractString)

Compute the checksum of the line `str` modulo 10.

The algorithm is simple: add all the numbers in the line, ignoring letters,
spaces, periods, and plus signs, but assigning +1 to the minus signs. The
checksum is the remainder of the division by 10.

"""
function compute_checksum(str::AbstractString)
    checksum = 0

    for c in str
        # Check if `c` is a number.
        if isnumeric(c)
            checksum += parse(Int, c)

        # Check if `c` is a minus sign, which has value 1.
        elseif c == '-'
            checksum += 1
        end

        # Otherwise, the value of the character is 0.
    end

    # Return the checksum modulo 10.
    checksum % 10
end

"""
    _show_tle(io::IO, tle::TLE, color::Bool = true)

Show the TLE `tle` in the IO `io`.

If `color` is `true`, then the text will be printed using colors. If `color` is
omitted, then it defaults to `true`.

"""
function _show_tle(io::IO, tle::TLE, color::Bool = true)
    # Colors will be printed only for STDOUT.
    b = (color) ? _b : ""
    d = (color) ? _d : ""
    g = (color) ? _g : ""
    y = (color) ? _y : ""

    # Auxiliary variables.
    name            = tle.name
    sat_num         = tle.sat_num
    int_designator  = tle.int_designator
    epoch_year      = tle.epoch_year
    epoch_day       = tle.epoch_day
    epoch           = tle.epoch
    epoch_str       = string(julian2datetime(epoch))
    elem_set_number = tle.elem_set_number
    e               = tle.e
    i               = tle.i
    Ω               = tle.Ω
    ω               = tle.ω
    M               = tle.M
    n               = tle.n
    rev_num         = tle.rev_num
    bstar           = tle.bstar
    dn_o2           = tle.dn_o2
    ddn_o6          = tle.ddn_o6

    # Print the TLE information.
    print(io, "TLE:\n")
    print(io, "$(b)                     Name : $(d)"); @printf(io, "%s\n", name)
    print(io, "$(b)         Satellite number : $(d)"); @printf(io, "%d\n", sat_num)
    print(io, "$(b) International designator : $(d)"); @printf(io, "%s\n", int_designator)
    print(io, "$(b)       Epoch (Year / Day) : $(d)"); @printf(io, "%d / %12.8f\n", epoch_year, epoch_day)
    print(io, "$(b)       Epoch (Julian Day) : $(d)"); @printf(io, "%12.5f (%s)\n", epoch, epoch_str)
    print(io, "$(b)       Element set number : $(d)"); @printf(io, "%d\n", elem_set_number)
    print(io, "$(b)             Eccentricity : $(d)"); @printf(io, "%12.8f deg\n", e)
    print(io, "$(b)              Inclination : $(d)"); @printf(io, "%12.8f deg\n", i)
    print(io, "$(b)                     RAAN : $(d)"); @printf(io, "%12.8f deg\n", Ω)
    print(io, "$(b)      Argument of perigee : $(d)"); @printf(io, "%12.8f deg\n", ω)
    print(io, "$(b)             Mean anomaly : $(d)"); @printf(io, "%12.8f deg\n", M)
    print(io, "$(b)          Mean motion (n) : $(d)"); @printf(io, "%12.8f revs/day\n", n)
    print(io, "$(b)        Revolution number : $(d)"); @printf(io, "%d\n", rev_num)
    print(io, "$(b)                       B* : $(d)"); @printf(io, "%f 1/[er]\n", bstar)
    print(io, "$(b)                    ṅ / 2 : $(d)"); @printf(io, "%f rev/day²\n", dn_o2)
    print(io, "$(b)                    n̈ / 6 : $(d)"); @printf(io, "%f rev/day³", ddn_o6)

    nothing
end

